# -*- coding: utf-8 -*-
"""
2018-07-20
@author: Maxime Beau, Neural Computations Lab, University College London
"""

import os
import os.path as op; opj=op.join
import psutil
from pathlib import Path
from tqdm.notebook import tqdm

from collections.abc import Iterable

import multiprocessing
num_cores = multiprocessing.cpu_count()

import numpy as np
from math import ceil

import matplotlib.pyplot as plt

from npyx.utils import npa, split, xcorr_1d_loop
from npyx.inout import read_metadata, get_binary_file_path, chan_map, whitening, bandpass_filter, apply_filter, med_substract
from npyx.gl import get_units, get_npyx_memory

def wvf(dp, u=None, n_waveforms=100, t_waveforms=82, selection='regular', periods='all',
        spike_ids=None, wvf_batch_size=10, ignore_nwvf=True,
        save=True, verbose=False, again=False,
        whiten=False, med_sub=False, hpfilt=False, hpfiltf=300,
        nRangeWhiten=None, nRangeMedSub=None, ignore_ks_chanfilt=True):
    '''
    ********
    routine from rtn.npyx.spk_wvf
    Extracts a sample of waveforms from the raw data file.
    ********

    Parameters:
        - dp:                 str or PosixPath, path to kilosorted dataset.
        - u:                  int, unit index.
        - n_waveforms:        int, number of waveform to return, selected according to the periods parameter | Default 100
        - t_waveforms:        int, temporal span of waveforms | Default 82 (about 3ms)
        - selection:          str, way to select subset of n_waveforms spikes to return the waveforms of.
                              Either 'regular' (homogeneous selection or in batches) or 'random'.
        - periods:            recording periods to sample waveforms from. Either 'all' (default)
                              or [(t1, t2), (t3, t4), ...] with t1, t2 in seconds.
        - spike_ids:          list/array, relative indices of spikes in the whole recording.
                                          Takes precedence over every other parameter: if provided, u, n_waveforms and periods will be ignored.
        - wvf_batch_size:     int, if >1 and 'regular' selection, selects ids as batches of spikes. | Default 10
        - save:               bool, whether to save to routine memory. | Default True
        - verbose:            bool, whether to print informaiton. | Default False
        - again:              bool, whether to recompute waveforms even if ofund in routines memory. | Default False
        - ignore_nwvf:        bool, whether to ignore n_waveforms parameter when a list of times is provided as periods,
                                    to return all the spikes in the window instead. | Default True
        - whiten:             bool, whether to whiten across channels.
                                    Globally by default, using the nRangeWhiten closest channels if nRangeWhiten is provided. | Default False
        - med_sub:            bool, whether to median-subtract across channels.
                                    Globally by default, using the nRangeMedSub closest channels if nRangeWhiten is provided. | Default False
        - hpfilt:             bool, whether to high-pass filter with a butterworth filter (order 3) of cutoff frequency hpfiltf. | Default False
        - hpfiltf:            int, high-pass filter cutoff frequency | Default 300
        - nRangeWhiten        int, number of channels to use to compute the local median. | Default None
        - nRangeMedSub:       int, number of channels to use to compute the local median. | Default None
        - ignore_ks_chanfilt: bool, whether to ignore kilosort channel filtering
                                    (if False, output shape will always be n_waveforms x t_waveforms x 384) | Default False
    Returns:
        waveforms:            numpy array of shape (n_waveforms x t_waveforms x n_channels)
                                    where n_channels is defined by the channel map if ignore_ks_chanfilt is False.

    '''
    dp = Path(dp)

    if spike_ids is not None:
        if u is not None and verbose: print('WARNING you provided both u and spike_ids! u is ignored.')
        if n_waveforms !=100 and verbose: print('WARNING you provided both n_waveforms and spike_ids! n_waveforms is ignored.')
        if not isinstance(periods,str) and verbose: print('WARNING you provided both periods and spike_ids! periods is ignored.')
        u=np.unique(np.load(Path(dp)/'spike_clusters.npy')[spike_ids])
        assert len(u)==1, 'WARNING the spike ids that you provided seem to belong to different units!! Double check!'
        u=u[0]
    dp, u = get_source_dp_u(dp, u)

    dpnm = get_npyx_memory(dp)

    if isinstance(periods, str): assert periods=='all', "WARNING periods should either be 'all' or [[t1,t2],[t3,t4]...]."
    per_str = str(periods)[0:50].replace(' ', '').replace('\n','')
    fn=f"wvf{u}_{n_waveforms}-{t_waveforms}_{per_str}_{hpfilt}{hpfiltf}-{whiten}{nRangeWhiten}-{med_sub}{nRangeMedSub}-{ignore_ks_chanfilt}.npy"
    if os.path.exists(Path(dpnm,fn)) and (not again) and (spike_ids is None):
        if verbose: print("File {} found in routines memory.".format(fn))
        return np.load(Path(dpnm,fn))

    waveforms = get_waveforms(dp, u, n_waveforms, t_waveforms, selection, periods, spike_ids, wvf_batch_size, ignore_nwvf,
                 whiten, med_sub, hpfilt, hpfiltf, nRangeWhiten, nRangeMedSub, ignore_ks_chanfilt, verbose)
    # Save it
    if (save and (spike_ids is None)):
        np.save(Path(dpnm,fn), waveforms)

    return waveforms

# def get_w(traces, slc, _n_samples_extract):
#     # Get slice
#     extract = traces[slc].astype(np.float32)
#     # Center channels individually
#     extract = extract-np.median(extract, axis=0)
#     # Pad the extracted chunk if at recording limit.
#     if slc.start <= 0: extract = _pad(extract, _n_samples_extract, 'left')
#     elif slc.stop >= traces.shape[0] - 1: extract = _pad(extract, _n_samples_extract, 'right')
#     # Add this waveform, all good!
#     return extract.T

def get_waveforms(dp, u, n_waveforms=100, t_waveforms=82, selection='regular', periods='all',
                  spike_ids=None, wvf_batch_size=10, ignore_nwvf=True,
                  whiten=0, med_sub=0, hpfilt=0, hpfiltf=300,
                  nRangeWhiten=None, nRangeMedSub=None, ignore_ks_chanfilt=0, verbose=False):
    f"{wvf.__doc__}"

    # Extract and process metadata
    dp = Path(dp)
    meta = read_metadata(dp)
    dat_path = get_binary_file_path(dp, 'ap')

    dp_source = get_source_dp_u(dp, u)[0]
    meta=read_metadata(dp_source)
    dtype=np.dtype(meta['highpass']['datatype'])
    n_channels_dat=meta['highpass']['n_channels_binaryfile']
    n_channels_rec = n_channels_dat-1 if meta['acquisition_software']=='SpikeGLX' else n_channels_dat
    sample_rate=meta['highpass']['sampling_rate']
    item_size = dtype.itemsize
    fileSizeBytes=meta['highpass']['binary_byte_size']
    if meta['acquisition_software']=='SpikeGLX':
        if meta['highpass']['fileSizeBytes'] != fileSizeBytes:
            print((f"\033[91;1mMismatch between ap.meta and ap.bin file size"
            "(assumed encoding is {str(dtype)} and Nchannels is {n_channels_dat})!! "
            f"Probably wrong meta file - just edit fileSizeBytes in the .ap.meta file at {dp} "
            f"(replace {int(meta['highpass']['fileSizeBytes'])} with {fileSizeBytes}) "
            "and be aware that something went wrong in your data management...\033[0m"))

    # Select subset of spikes
    spike_samples = np.load(Path(dp, 'spike_times.npy'), mmap_mode='r').squeeze()
    if spike_ids is None:
        spike_ids_subset=get_ids_subset(dp, u, n_waveforms, wvf_batch_size, selection, periods, ignore_nwvf, verbose)
    else:
        assert isinstance(spike_ids, Iterable), "WARNING spike_ids must be a list/array of ids!"
        spike_ids_subset=np.array(spike_ids)
    n_spikes = len(spike_ids_subset)

    # Get waveforms times in bytes
    # and check that, for this waveform width,
    # they no not go beyond file limits
    waveforms_t = spike_samples[spike_ids_subset].astype(np.int64)
    waveforms_t1 = (waveforms_t-t_waveforms//2)*n_channels_dat*item_size
    waveforms_t2 = (waveforms_t+t_waveforms//2)*n_channels_dat*item_size
    wcheck_m=(0<=waveforms_t1)&(waveforms_t2<fileSizeBytes)
    if not np.all(wcheck_m):
        print(f"Invalid times: {waveforms_t[~wcheck_m]}")
        waveforms_t1 = waveforms_t1[wcheck_m]
        waveforms_t2 = waveforms_t2[wcheck_m]

    # Iterate over waveforms
    waveforms = np.zeros((n_spikes, t_waveforms, n_channels_rec), dtype=np.float32)
    if verbose: print(f'Loading waveforms of unit {u} ({n_spikes})...')
    with open(dat_path, "rb") as f:
        for i,t1 in enumerate(waveforms_t1):
            if n_spikes>10:
                if i%(n_spikes//10)==0 and verbose: print(f'{round((i/n_spikes)*100)}%...', end=' ')
            f.seek(t1, 0) # 0 for absolute file positioning
            wave=f.read(n_channels_dat*t_waveforms*item_size)
            wave=np.frombuffer(wave, dtype=dtype).reshape((t_waveforms,n_channels_dat))
            wave = wave-np.median(wave, axis = 0)[np.newaxis,:] # center the waveforms on 0
            # get rid of sync channel
            waveforms[i,:,:] = wave[:,:-1] if meta['acquisition_software']=='SpikeGLX' else wave
    if verbose: print('\n')

    # Preprocess waveforms
    if hpfilt|med_sub|whiten:
        waveforms=waveforms.reshape((n_spikes*t_waveforms, n_channels_rec))
        if hpfilt:
            waveforms=apply_filter(waveforms, bandpass_filter(rate=sample_rate, low=None, high=hpfiltf, order=3), axis=0)
        if med_sub:
            waveforms=med_substract(waveforms, axis=1, nRange=nRangeMedSub)
        if whiten:
            waveforms=whitening(waveforms.T, nRange=nRangeWhiten).T # whitens across channels so gotta transpose
        waveforms=waveforms.reshape((n_spikes,t_waveforms, n_channels_rec))

    # Filter channels ignored by kilosort if necesssary
    if not ignore_ks_chanfilt:
        channel_ids_ks = np.load(Path(dp, 'channel_map.npy'), mmap_mode='r').squeeze()
        channel_ids_ks=channel_ids_ks[channel_ids_ks!=384]
        waveforms=waveforms[:,:,channel_ids_ks] # only AFTER processing, filter out channels

    # Correct voltage scaling
    waveforms*=meta['bit_uV_conv_factor']

    return  waveforms.astype(np.float32)

def wvf_dsmatch(dp, u, n_waveforms=100, t_waveforms=82, periods='all',
                wvf_batch_size=10, ignore_nwvf=True, med_sub = False, spike_ids = None,
                save=True, verbose=False, again=False,
                whiten=False,  hpfilt=False, hpfiltf=300, nRangeWhiten=None, nRangeMedSub=None,
                n_waves_used_for_matching = 50000, peakchan_allowed_range=10,
                use_average_peakchan = False, max_allowed_amplitude = 1800, max_allowed_shift=3,
                n_waves_to_average=5000, plot_debug=False, do_shift_match=True, n_waveforms_per_batch=10):
    """
    ********
    Extract the drift and shift matched mean waveforms of the specified unit.
    Drift and shift matching consists of two steps:

    First: load all waveforms, average them 10 by 10 = 'spike batches'
    (else, they would be too noisy to work with. Assumption: 10 consecutive waveforms have the same 'drift state')

    Drift matching:
        - Z-drift-matching: sub-select spike batches peaking on same peak channel (modal channel)
        - XY-drift-marching: sub-select n_waves_to_average/10 spikes batches
          with the highest amplitude on this peak channel
          (up to 99th percentile, not highest amplitude)
        - Exclude batches with amplitude higher than max_allowed_amplitude uV (gets rid of potential artefacts)

    Shift matching:
        - Define a template from the 5 drift-matched batches with the highest amplitude
        - Compute crosscorrelation between each batch and template
        - Re-align each batch to the template accordingly to peaking crosscorrelation
        - Exclude batches which were required to shift by more than +/-max_allowed_shift samples
          (naturally gets rid of
          - completely off batches - there is probably something wrong with them
          - noisy batches - which do not match the template well enough for a match to be found around 0

    To diagnose issues: set **plot_debug=True** (and again=true of course), to plot the distributions of peak channel, amplitudes and shifts

    Currently only supports passing a single unit as input, hence
    prints error message if 'spike_ids = single_slice' if passed.
    ********


     Parameters:
        - dp:                 str or PosixPath, path to kilosorted dataset.
        - u:                  int, unit index.
        - n_waveforms:        int, number of waveform to return, selected according to the periods parameter | Default 100
        - t_waveforms:        int, temporal span of waveforms | Default 82 (about 3ms)
        - periods:   str/list of tuples, either 'regular' (homogeneous selection or in batches), 'random',
                                                  or a list of time chunks [(t1, t2), (t3, t4), ...] with t1, t2 in seconds.
        - spike_ids:          list/array, relative indices of spikes in the whole recording.
                                          If provided, u, n_waveforms and periods will be ignored.
        - wvf_batch_size:     int, if >1 and 'regular' selection, selects ids as batches of spikes. | Default 10
        - save: bool,         whether to save to routine memory. | Default True
        - verbose: bool,         whether to print information. | Default False
        - again: bool,        whether to recompute waveforms even if found in routines memory. | Default False
        - ignore_nwvf:        bool, whether to ignore n_waveforms parameter when a list of times is provided as periods,
                                    to return all the spikes in the window instead. | Default True
        - whiten:             bool, whether to whiten across channels.
                                    Globally by default, using the nRangeWhiten closest channels if nRangeWhiten is provided. | Default False
        - med_sub:            bool, whether to median-subtract across channels.
                                    Globally by default, using the nRangeMedSub closest channels if nRangeWhiten is provided. | Default False
        - hpfilt:             bool, whether to high-pass filter with a butterworth filter (order 3) of cutoff frequency hpfiltf. | Default False
        - hpfiltf:            int, high-pass filter cutoff frequency | Default 300
        - nRangeWhiten        int, number of channels to use to compute the local median. | Default None
        - nRangeMedSub:       int, number of channels to use to compute the local median. | Default None
        - ignore_ks_chanfilt: bool, whether to ignore kilosort channel filtering
                                    (if False, output shape will always be n_waveforms x t_waveforms x 384) | Default False
        - n_waves_used_for_matching:   int, how many spikes to subsample to perform matching (default 50000 waveforms)
        - peakchan_allowed_range: int (channel id), maximum allowed distance between original pek channel and ds-matched drift channel
        - use_average_peakchan: bool, if True simply use the channel with highest average amplitude across spikes as peak channel
                                       instead of using the channel where the most spikes peak on
        - max_allowed_amplitude: float, maximum amplitude in uV (peak to trough) that a spike average can have to be considered (above, must be artefactual)
        - max_allowed_shift: int (samples), maximum allowed temporal shift during shift-matching (see Shift-matching explanation above)
        - n_waves_to_average: int, maximum number of waveforms averaged together (5000 waveforms = 500 batches)
        - plot_debug: bool, whether to plot informative histograms displaying the distribution of peak channels (Z drift matching),
                      amplitudes on this peak channel (XY drift matching) and shifts (shift matching)
        - do_shift_match: bool, whether to perform shift matching
        - n_waveforms_per_batch: int, number of waveforms to use per batch for drift matching
                                 (in an ideal world 1, but too noisy - we assume that
                                  n_waveforms_per_batch consecutive waveforms have the same drift state)

    Returns:
        - peak_dsmatched_waveform: (n_samples,) array (t_waveforms samples) storing the peak channel waveform
        - dsmatched_waveform: (n_samples, n_channels) array storing the drift-shift-matched waveform across channels (384 for Neuropixels 1.0)
        - spike_ids: (n_spikes,) array of absolute ids (w/r all spikes in recording)
                     of spikes subset selected to compute the final drift-shift-matched waveform
        - peak_channel: (1,) array storing the channel used to select the subset of waveforms during drift matching (de facto, peak channel)
    """

    dp = Path(dp)

    if spike_ids is not None:
        raise ValueError('No support yet for passing multiple spike indices. Exiting.')


    dpnm = get_npyx_memory(dp)

    per_str = str(periods)[0:50].replace(' ', '').replace('\n','')
    fn=f"dsm_{u}_{n_waveforms}-{t_waveforms}_{per_str}_{hpfilt}{hpfiltf}-{whiten}{nRangeWhiten}-{med_sub}{nRangeMedSub}.npy"
    fn_all=f"dsm_{u}_all_waves_{n_waveforms}-{t_waveforms}_{per_str}_{hpfilt}{hpfiltf}-{whiten}{nRangeWhiten}-{med_sub}{nRangeMedSub}.npy"
    fn_spike_id=f"dsm_{u}_spike_id_{n_waveforms}-{t_waveforms}_{per_str}_{hpfilt}{hpfiltf}-{whiten}{nRangeWhiten}-{med_sub}{nRangeMedSub}.npy"
    fn_peakchan=f"dsm_{u}_peakchan_{n_waveforms}-{t_waveforms}_{per_str}_{hpfilt}{hpfiltf}-{whiten}{nRangeWhiten}-{med_sub}{nRangeMedSub}.npy"

    if Path(dpnm,fn).is_file() and (not again) and (spike_ids is None):
        if verbose: print(f"File {fn} found in routines memory.")
        drift_shift_matched_mean = np.load(Path(dpnm,fn_all))
        if plot_debug:
            w=wvf(dp, u, n_waveforms=n_waveforms, t_waveforms=t_waveforms)
            fig = quickplot_n_waves(np.mean(w, 0))
            fig = quickplot_n_waves(drift_shift_matched_mean, f'blue: 100 random waveforms\norange: dsmatched_waveforms (unit {u})', fig=fig)
        return np.load(Path(dpnm,fn)),drift_shift_matched_mean,np.load(Path(dpnm,fn_spike_id)), np.load(Path(dpnm,fn_peakchan))

    ## Extract spike ids so we can extract consecutive waveforms
    spike_ids_all = ids(dp, u, periods=periods)
    spike_ids_split_all = split(spike_ids_all, n_waveforms_per_batch, return_last = False).astype(np.int64)
    
    ## Subsample waveforms based on available RAM
    vmem=dict(psutil.virtual_memory()._asdict())
    available_RAM = vmem['total']-vmem['used']
    single_w_size = wvf(dp, None, t_waveforms=t_waveforms, spike_ids=[0]).nbytes
    max_n_waveforms = available_RAM//single_w_size-100 # -100 to be safe
    n_waves_used_for_matching = min(n_waves_used_for_matching, max_n_waveforms)
    
    # now subsample ids based on the RAM-safe n_waves_used_for_matching
    if n_waves_used_for_matching<len(spike_ids_all):
        n_batches_used_for_matching=n_waves_used_for_matching//n_waveforms_per_batch # floor division
        spike_ids_subsample=np.round(np.linspace(0,spike_ids_split_all.shape[0]-1,n_batches_used_for_matching)).astype(int)
        spike_ids_split = spike_ids_split_all[spike_ids_subsample]
    else:
        spike_ids_split=spike_ids_split_all
    spike_ids_split_indices = np.arange(0,spike_ids_split.shape[0],1)

    ## Extract the waveforms using the wvf function in blocks of 10 (n_waveforms_per_batch).
    # After waves have been extracted, put the index of the channel with the
    # max amplitude as well as the max amplitude into the peak_chan_split array
    spike_ids_split = spike_ids_split.flatten()
    raw_waves = wvf(dp, u = None, n_waveforms= 100, t_waveforms = t_waveforms,
                    selection='regular', periods=periods, spike_ids=spike_ids_split,
                    wvf_batch_size =wvf_batch_size , ignore_nwvf=ignore_nwvf,
                    save=save , verbose = verbose,  again=True, whiten = whiten,
                    hpfilt = hpfilt, hpfiltf = hpfiltf, nRangeWhiten=nRangeWhiten,
                    nRangeMedSub=nRangeMedSub, ignore_ks_chanfilt=True)
    spike_ids_split = spike_ids_split.reshape(-1,n_waveforms_per_batch)
    raw_waves = raw_waves.reshape(spike_ids_split.shape[0], n_waveforms_per_batch, t_waveforms, -1)
    mean_waves = np.mean(raw_waves, axis = 1)
    ## Find peak channel (and store amplitude) of every batch
    # only consider amplitudes on channels around original peak channel
    original_peak_chan = get_peak_chan(dp, u)
    c_left, c_right = max(0, original_peak_chan-peakchan_allowed_range), min(original_peak_chan+peakchan_allowed_range, mean_waves.shape[2])
    # calculate amplitudes ("peak-to-peak"), but ONLY using 2ms (-30,30) in the middle
    t1, t2 = max(0,mean_waves.shape[1]//2-30), min(mean_waves.shape[1]//2+30, mean_waves.shape[1])
    amplitudes = np.ptp(mean_waves[:,t1:t2,c_left:c_right], axis=1)
    batch_peak_channels = np.zeros(shape=(spike_ids_split_indices.shape[0], 3))
    batch_peak_channels[:,0] = spike_ids_split_indices # store batch indices (batch = averaged 10 spikes)
    batch_peak_channels[:,1] = c_left+np.argmax(amplitudes, axis = 1) # store peak channel of each batch
    batch_peak_channels[:,2] = np.max(amplitudes, axis = 1) # store peak channel amplitude

    # Filter out batches with too large amplitude (probably artefactual)
    batch_peak_channels = batch_peak_channels[batch_peak_channels[:,2] < max_allowed_amplitude]

    #### Z-drift matching ####
    # subselect batches with same peak channel
    if use_average_peakchan:
        peak_channel = int(original_peak_chan)
    else:
        # use mode of peak channel distribution across spikes
        chans, count = np.unique(batch_peak_channels[:,1], return_counts = True)
        peak_channel = int(chans[np.argmax(count)])

    if plot_debug:
        fig = hist_MB(batch_peak_channels[:,1], a=peak_channel-20, b=peak_channel+20, s=1,
        title=f'Z drift matching:\ndistribution of peak channel across spike batches\n({n_waveforms_per_batch} spikes/batch - mode: chan {peak_channel})')
        ylim = fig.get_axes()[0].get_ylim()
        fig.get_axes()[0].plot([peak_channel,peak_channel], ylim, color='red', ls='--')
        fig.get_axes()[0].set_ylim(ylim)

    batch_peak_channels = batch_peak_channels[batch_peak_channels[:,1] == peak_channel]

    #### X-Y-drift matching ####
    # subselect batches with similar amplitude (i.e. similar distance to probe)
    # and in particular, close to largest amplitude (close to probe, but not max to avoid artefacts)
    # aim for 500 spikes (50 batches)
    # should average enough, but still use a small subset of drift-matched spikes!
    n_driftmatched_subset = n_waves_to_average//n_waveforms_per_batch
    batch_peak_channels = batch_peak_channels[np.argsort(batch_peak_channels[:,2])] # sort by amplitude

    if plot_debug:
        max_amp_hist = np.max(batch_peak_channels[:,2])
        max_amp_hist += 10-max_amp_hist%10
        nbatches_hist = batch_peak_channels.shape[0]
        fig = hist_MB(batch_peak_channels[:,2], a=10, b=max_amp_hist, s=5, color='grey', alpha=0.7)

    # if less than n_driftmatched_subset batches below 95th percentile,
    # use all up to n_driftmatched_subset batches
    ## TODO only perform the following if distribution is unimodal
    # (Hartigan Dip-test of Unimodality)
    prct_95_i = int(batch_peak_channels.shape[0]*0.95)
    if prct_95_i<n_driftmatched_subset: 
        batch_peak_channels = batch_peak_channels[0:n_driftmatched_subset]
    else:
        i_left = max(prct_95_i - n_driftmatched_subset, 0) # should never be negative given if statement, but precaution
        batch_peak_channels = batch_peak_channels[i_left:prct_95_i]
    drift_matched_spike_ids = np.sort(batch_peak_channels[:,0])

    if plot_debug:
        fig = hist_MB(batch_peak_channels[:,2], a=10, b=max_amp_hist, s=5, ax=fig.get_axes()[0], color='orange', alpha=0.7,
        title=(f'XY drift matching:\ndistribution of amplitude on peak channel across spike batches\n'
               f'({n_waveforms_per_batch} spikes/batch - {batch_peak_channels.shape[0]}/{nbatches_hist} batches)'))


    #### shift matching ####
    # extract drift-matched raw waveforms
    dsmatch_batch_ids = batch_peak_channels[:,0].astype(np.int64)
    drift_matched_waves = raw_waves[dsmatch_batch_ids]#.reshape(-1, t_waveforms, raw_waves.shape[-1])
    drift_matched_batches = np.mean(drift_matched_waves, axis=1)

    # shift waves using simple negative peak matching
    recenter_spikes = False
    if do_shift_match:
        drift_shift_matched_batches = shift_match(drift_matched_batches, peak_channel, max_allowed_shift, recenter_spikes, plot_debug)
    else:
        drift_shift_matched_batches = drift_matched_batches
    # Get the median of the drift and shift matched waves (not sensitive to outliers)
    drift_shift_matched_mean = np.median(drift_shift_matched_batches, axis=0)
    drift_shift_matched_mean_peak = drift_shift_matched_mean[:,peak_channel]
    # recenter spike absolute maximum
    shift = (np.argmax(np.abs(drift_shift_matched_mean_peak)) - drift_shift_matched_mean_peak.shape[0]//2)%drift_shift_matched_mean_peak.shape[0]
    drift_shift_matched_mean = np.concatenate([drift_shift_matched_mean[shift:], drift_shift_matched_mean[:shift]], axis=0)
    drift_shift_matched_mean_peak = np.concatenate([drift_shift_matched_mean_peak[shift:], drift_shift_matched_mean_peak[:shift]], axis=0)

    if save:
        np.save(Path(dpnm,fn), drift_shift_matched_mean_peak)
        np.save(Path(dpnm,fn_all), drift_shift_matched_mean)
        np.save(Path(dpnm,fn_spike_id), drift_matched_spike_ids)
        np.save(Path(dpnm, fn_peakchan), peak_channel)

    if plot_debug:
        if verbose: print(f'Total averaged waveform batches ({n_waveforms_per_batch}/batch) after drift-shift matching: {batch_peak_channels.shape[0]}')
        fig = quickplot_n_waves(np.mean(mean_waves[np.random.randint(0, mean_waves.shape[0], batch_peak_channels.shape[0]),:,:], axis=0), '', peak_channel)
        fig = quickplot_n_waves(np.mean(drift_matched_batches, axis=0), '', peak_channel, fig=fig)
        fig = quickplot_n_waves(drift_shift_matched_mean, 'raw:blue\ndrift-matched:orange\ndrift-shift-matched:green', peak_channel, fig=fig)
        #breakpoint()

    return drift_shift_matched_mean_peak, drift_shift_matched_mean, drift_matched_spike_ids, peak_channel

def shift_match(waves, alignment_channel,
                chan_range=2, recenter_spikes=False,
                plot_debug=False, dynamic_template=False,
                max_shift_allowed = 5):
    """
    Iterate through waveforms to align them to each other
    by maximizing their convolution.

    In order to have a time complexity O(n),
    starts by aligning waves[1] to waves[0],
    then waves[2] to mean(waves[[0], aligned waves[1]])...
    So every wave will be aligned to the first wave.

    When shifting a wave, fills the gap
    with the bit clipped from the other end.

    Parameters:
        - waves: (n_waves, n_samples, n_channels,) array
        - alignment_channel: channel to use to compute the convolution
        - chan_range: int, range of channels around alignment channel used for template matching
                      (3 corresponds to 6 channels, 5 to 10 etc)
        - recenter_spikes: bool, whether to align the maximum of template to 0
        - dynamic_template: bool, whether to update the template by averaging it with the aligned spike
        - max_shift_allowed: int, maximum shift allowed (half window) - other waveforms are discarded (if need to shift more, they must be way too noisy)

    Returns:
        - shifted_waves: (n_waves, n_samples, n_channels,) array
    """
    

    # sort waveforms by amplitude
    amplitudes = np.ptp(waves[:,:,alignment_channel], axis=1)
    amplitudes_i = np.argsort(amplitudes, axis=0)
    waves_sort = waves[amplitudes_i[::-1],:,:]
    # use median of 50 waves of highest amplitude as template
    # most arbitrary decision 0 but seems reasonable and empirically works
    n_waveforms_template=50
    template = np.median(waves_sort[:n_waveforms_template,:,:], axis=0)
    if recenter_spikes:
        shift = (np.argmax(np.abs(template[:,alignment_channel])) - template.shape[0]//2)%template.shape[0]
        if plot_debug:
            plt.figure()
            plt.plot(template[:,alignment_channel])
        template = np.concatenate([template[shift:], template[:shift]], axis=0)# shift template to center maximum
        if plot_debug: plt.plot(template[:,alignment_channel])

    # initialize array
    aligned_waves = np.zeros(waves_sort.shape)
    chan_min, chan_max = max(0,alignment_channel-chan_range), min(alignment_channel+chan_range, template.shape[1])
    template = template[:,chan_min:chan_max] # defined across 10 closest channels
    shifts = []
    for i, w in enumerate(waves_sort):
        w_closestchannels = w[:,chan_min:chan_max]
        xcorr_w_template = xcorr_1d_loop(template, w_closestchannels)
        # average xcorr across channels to find the optimal alignment
        # using information from all channels around peak!
        xcorr_max = np.argmax(np.mean(xcorr_w_template, axis=1))
        shift = (xcorr_max-waves_sort.shape[1]//2)%waves_sort.shape[1]
        relative_shift = (shift+waves_sort.shape[1]//2)%waves_sort.shape[1]-waves_sort.shape[1]//2
        shifts.append(relative_shift)
        realigned_w = np.concatenate([w[-shift:,:], w[:-shift,:]], axis=0)
        if abs(relative_shift)>max_shift_allowed:
            realigned_w=realigned_w*np.nan
        if plot_debug and i==0:
            fig=imshow_cbar(template)
            fig=imshow_cbar(w_closestchannels)
            fig=imshow_cbar(xcorr_w_template)
        # store realigned_wave in array
        aligned_waves[i,:,:] = realigned_w

        # optionnally update template by averaging with realigned waveform
        # didn't work very well - better to keep a relatively 'focused' template
        if dynamic_template:
            template = np.mean(np.stack(
                            [template,
                            realigned_w[:,chan_min:chan_max]],
                            axis=2), axis=2)

    # discard nans (beyond max_shift_allowed) and re-sort waves properly
    aligned_waves = aligned_waves[np.arange(aligned_waves.shape[0])[amplitudes_i[::-1]],:,:]
    nan_m = np.isnan(aligned_waves[:,0,0])
    aligned_waves = aligned_waves[~nan_m,:,:]

    if plot_debug:
        #fig = imshow_cbar(template.T)
        a = np.max(np.abs(shifts))
        a += 10-a%10
        fig = hist_MB(shifts, s=1, a=-a, b=a, color='darkgreen', alpha=1,
        title=(f'Shift matching:\ndistribution of shifts w/r template across spike batches'))
        ylim = fig.get_axes()[0].get_ylim()
        fig.get_axes()[0].plot([max_shift_allowed,max_shift_allowed], ylim, color='red', ls='--')
        fig.get_axes()[0].plot([-max_shift_allowed,-max_shift_allowed], ylim, color='red', ls='--')
        fig.get_axes()[0].set_ylim(ylim)

    return aligned_waves

def get_pc(waveforms):
    wvf_m = np.mean(waveforms, axis=0)
    max_min_wvf=np.max(wvf_m,0)-np.min(wvf_m,0)
    peak_chan = np.argmax(max_min_wvf)
    return peak_chan

def get_peak_chan(dp, unit, use_template=True, again=False, ignore_ks_chanfilt=True, periods='all'):
    '''
    Returns index of peak channel, either according to the full probe channel map (0 through 383)
                                   or according to the kilosort channel map (0 through N with N<=383)

    Parameters:
        - datapath, string
        - unit, integer or float (if merged dataset)
        - use_template: bool, whether to use templates instead of raw waveform to find peak channel.
        - again: whether to recompute the waveforms/templates
        - ignore_ks_chanfilt: bool, whether to return the channel rank on the full probe
                    rather than the channel rank on the kilosort cahnnel map (jumping some channels).
                    They will be the same if all channels are used for spike sorting.
                    E.g. if kilosort only used 380 channels (out of 384),
                    the last channel, 383, has the relative index 379.
    Returns:
        - best_channel, integer indexing the channel
          where the unit averaged raw waveform (n=100 spanning the whole recording)
          has the largest peak to trough amplitude.

          WARNING: this value is ABSOLUTE ON THE PROBE SHANK BY DEFAULT. If you wish the relative channel index
          taking into account channels ignored by kilosort, set ignore_ks_chanfilt to False.
    '''
    dp = Path(dp)
    dp, unit = get_source_dp_u(dp, unit)

    strdic={True:'templates', False:'raw-waveforms'}
    f_all=f'peak_channels_{strdic[use_template]}_all.npy'
    f_good=f'peak_channels_{strdic[use_template]}_good.npy'
    for f in [f_all, f_good]:
        if op.exists(Path(dp, f)):
            peak_chans=np.load(Path(dp, f))
            if unit in peak_chans[:,0]:
                return int(peak_chans[peak_chans[:,0]==unit, 1])

    cm=chan_map(dp, probe_version='local')
    if use_template:
        waveforms=templates(dp, unit)
        ks_peak_chan = get_pc(waveforms)
        peak_chan = cm[:,0][ks_peak_chan]
    else:
        waveforms=wvf(dp, u=unit, n_waveforms=200, t_waveforms=82,
                      selection='regular', periods=periods, spike_ids=None, again=again,
                      ignore_ks_chanfilt=True)
        probe_peak_chan = get_pc(waveforms)
        if ignore_ks_chanfilt: # absolute == relative channel index
            peak_chan = probe_peak_chan
        else: #
            ks_peak_chan = np.nonzero(cm[:,0]==probe_peak_chan)
            peak_chan = ks_peak_chan

    return int(peak_chan)


def get_depthSort_peakChans(dp, units=[], quality='all', use_template=True, again=False, verbose = False):
    '''
    Usage:
        Either feed in a list of units - the function will return their indices/channels sorted by depth in a n_units x 2 array,
        or simply indicate units 'quality' as 'all', 'mua' or good - will behave as if you had fed the list of units of this given quality.
    Parameters:
        - datapath, string
        - units, list of integers or strings
        - quality: string, 'all', 'mua' or 'good'
    Returns:
        - best_channels, numpy array of shape (n_units, 2).
          Column 1: unit indices, column 2: respective peak channel indices.
    '''

    dp = Path(dp)
    save=False # can only turn True if no (i.e. all) units are fed
    strdic={True:'templates', False:'raw-waveforms'}

    if len(units)==0:
        # If no units, load them all from dataset
        # and prepare to save the FULL array of peak channels at the end
        units=get_units(dp, quality=quality, again=again)
        assert np.any(units), f'No units of quality {quality} found in this dataset.'
        pc_fname=f'peak_channels_{strdic[use_template]}_{quality}.npy'
        if op.exists(Path(dp, pc_fname)) and not again:
            peak_chans=np.load(Path(dp, pc_fname))
            if np.all(np.isin(units, peak_chans[:,0])):
                return peak_chans
            else:
                save=True
        else:
            save=True
    else:
        # If units are fed, try to load the peak channels
        # from the saved FULL array of peak channels
        units=npa(units).flatten()
        f_all=f'peak_channels_{strdic[use_template]}_all.npy'
        f_good=f'peak_channels_{strdic[use_template]}_good.npy'
        for f in [f_all, f_good]:
            if op.exists(Path(dp, f)):
                peak_chans=np.load(Path(dp, f))
                if np.all(np.isin(units, peak_chans[:,0])):
                    units_mask=np.isin(peak_chans[:,0], units)
                    return peak_chans[units_mask]

    dt=np.float64 if assert_multi(dp) else np.int64
    peak_chans=npa(zeros=(len(units),2),dtype=dt)
    for iu, u in enumerate(units):
        if verbose: print("Getting peak channel of unit {}...".format(u))
        peak_chans[iu,0] = u
        peak_chans[iu,1] = np.array([get_peak_chan(dp, u, use_template)]).astype(dt)
    if assert_multi(dp):
        depth_ids = np.lexsort((-peak_chans[:,1], get_ds_ids(peak_chans[:,0])))
    else:
        depth_ids = np.argsort(peak_chans[:,1])[::-1] # From surface (high ch) to DCN (low ch)
    peak_chans=peak_chans[depth_ids,:]

    if save:
        np.save(Path(dp, pc_fname), peak_chans)

    return peak_chans # units, channels

def get_peak_pos(dp, unit, use_template=False, periods='all'):
    "Returns [x,y] relative position on the probe in um (y=0 at probe tip)."
    dp, unit = get_source_dp_u(dp, unit)
    peak_chan=get_peak_chan(dp, unit, use_template, periods=periods)
    pos = np.load(Path(dp,'channel_positions.npy'))
    cm=chan_map(dp, probe_version='local')
    return pos[cm[:,0]==peak_chan].ravel()

def get_chDis(dp, ch1, ch2):
    '''dp: datapath to dataset
    ch1, ch2: channel indices (1 to 384)
    returns distance in um.'''
    assert 1<=ch1<=384
    assert 1<=ch2<=384
    ch_pos = np.load(Path(dp,'channel_positions.npy')) # positions in um
    ch_pos1=ch_pos[ch1-1] # convert from ch number to ch relative index
    ch_pos2=ch_pos[ch2-1] # convert from ch number to ch relative index
    chDis=np.sqrt((ch_pos1[0]-ch_pos2[0])**2+(ch_pos1[1]-ch_pos2[1])**2)
    return chDis

def templates(dp, u, ignore_ks_chanfilt=False):
    '''
    ********
    routine from rtn.npyx.spk_wvf
    Extracts the template used by kilosort to cluster this unit.
    ********

    Parameters:
        - dp: string, datapath to kilosort output
        - unit: int, unit index
        - ignore_ks_chanfilt: bool, whether to ignore kilosort channel map (skipping unactive channels).
                              If true, n_channels is lower than 384.

    Returns:
        temaplate: numpy array of shape (n_templates, 82, n_channels) where n_channels is defined by the channel map and n_templates by how many merges were done.

    '''
    dp, u = get_source_dp_u(dp, u)

    IDs=ids(dp,u)
    #mean_amp=np.mean(np.load(Path(dp,'amplitudes.npy'))[IDs])
    template_ids=np.unique(np.load(Path(dp,'spike_templates.npy'))[IDs])
    templates = np.load(Path(dp, 'templates.npy'))[template_ids]#*mean_amp

    if ignore_ks_chanfilt:
        cm=chan_map(dp, y_orig='surface', probe_version='local')[:,0]
        cm_all=chan_map(dp, y_orig='surface', probe_version=None)[:,0]
        jumped_chans=np.sort(cm_all[~np.isin(cm_all, cm)])
        for ch in jumped_chans: # need to use a for loop because indices are right only absolutely, not relatively
            templates=np.insert(templates, ch, np.zeros(templates.shape[1]), 2)

    return templates

#%% wvf utilities

def get_ids_subset(dp, unit, n_waveforms, batch_size_waveforms, selection, periods, ignore_nwvf, verbose=False):
    
    # if periods were provided
    if not isinstance(periods, str):
        ids_subset = ids(dp, unit, periods=periods)
        if not ignore_nwvf:
            n_waveforms1=min(n_waveforms, len(ids_subset))
            ids_subset = np.unique(np.random.choice(ids_subset, n_waveforms1, replace=False))
        if verbose: print(f'In subset {periods}, {len(ids_subset)} waveforms were found (n_waveforms={n_waveforms}).')
    # if no periods were provided
    else:
        if n_waveforms in (None, 0):
            ids_subset = ids(dp, unit)
        else:
            assert n_waveforms > 0
            spike_ids = ids(dp, unit)
            assert any(spike_ids)
            assert selection in ['regular', 'random']
            if selection == 'regular':
                ids_subset = select_waveforms_in_batch(spike_ids, n_waveforms, batch_size_waveforms)
            elif selection == 'random' and len(spike_ids) > n_waveforms:
                ids_subset = np.unique(np.random.choice(spike_ids, n_waveforms, replace=False))

    return ids_subset

def select_waveforms_in_batch(spike_ids, n_waveforms, batch_size_waveforms):
    "Batch selection of spikes."
    if batch_size_waveforms is None or len(spike_ids) <= max(batch_size_waveforms, n_waveforms):
        step = ceil(np.clip(1. / n_waveforms * len(spike_ids),
                        1, len(spike_ids)))
        ids_subset = spike_ids[0::step][:n_waveforms]
    else:
        n_excerpts=n_waveforms // batch_size_waveforms
        excerpt_size=batch_size_waveforms
        ids_subset = np.concatenate([data_chunk(spike_ids, chunk)
                    for chunk in excerpts(len(spike_ids),
                                        n_excerpts=n_excerpts,
                                        excerpt_size=excerpt_size)])
        
    return ids_subset

def data_chunk(data, chunk, with_overlap=False):
    """Get a data chunk."""
    assert isinstance(chunk, tuple)
    if len(chunk) == 2:
        i, j = chunk
    elif len(chunk) == 4:
        if with_overlap:
            i, j = chunk[:2]
        else:
            i, j = chunk[2:]
    else:
        raise ValueError("'chunk' should have 2 or 4 elements, "
                         "not {0:d}".format(len(chunk)))
    return data[i:j, ...]

def excerpts(n_samples, n_excerpts=None, excerpt_size=None):
    """Yield (start, end) where start is included and end is excluded."""
    assert n_excerpts >= 2
    step = _excerpt_step(n_samples,
                         n_excerpts=n_excerpts,
                         excerpt_size=excerpt_size)
    for i in range(n_excerpts):
        start = i * step
        if start >= n_samples:
            break
        end = min(start + excerpt_size, n_samples)
        yield start, end

def _excerpt_step(n_samples, n_excerpts=None, excerpt_size=None):
    """Compute the step of an excerpt set as a function of the number
    of excerpts or their sizes."""
    assert n_excerpts >= 2
    step = max((n_samples - excerpt_size) // (n_excerpts - 1),
               excerpt_size)
    return step

# Recurrent imports
from npyx.merger import assert_multi, get_ds_ids, get_source_dp_u
from npyx.spk_t import ids
from npyx.plot import hist_MB, imshow_cbar, quickplot_n_waves



# Functions regarding information theoretic analysis of neural data

import numpy as np
from pyparsing import CaselessLiteral

from npyx.utils import smooth
from npyx.behav import get_processed_ifr


### average of pairwise synchronies i.e. P(a int b)/P(a)P(b)

def sync_wr_chance_shadmehr(t1, t2, events, binsize=2, window=[-500, 500],
                   remove_empty_trials=False, smooth_sd=4, pre_smooth=False, post_smooth=False,
                   y1=None, y2=None, return_trials = False, continuous=True, return_terms=False):
    """
    y1 and y2 should be T trials x B bins 2D matrices
    containing integer numbers of spikes (in theory only 0 and 1; any integer above 1 will be reduced to 1.)

    Relates to correlation as follows:
    Corr(X,Y) = Cov(X,Y)/std(X)std(Y)
              = Cov(X,Y)/sqrt(Var(X)Var(Y))
    Cov(X,Y) = E(XY) - E(X)E(Y)
             = P(X int Y) - P(X)P(Y)
    Var(X) = E(X**2) - E(X)**2
           = P(X) - P(X)P(X) because X is binary
           = P(X)(1-P(X))

    So Corr(X,Y) = (P(X int Y) - P(X)P(Y)) / sqrt(P(X)(1-P(X))P(Y)(1-P(Y)))
                 = (P(X int Y)/P(X)P(Y) - 1) / (sqrt(P(X)P(Y))*sqrt((1-P(X))(1-P(Y))) / P(X)P(Y))
                 = (sync(X,Y) - 1) / sqrt( (1-P(X))(1-P(Y)) / P(X)P(Y) )
                 = (sync(X,Y) - 1) / sqrt( 1/(1-P(X))(1-P(Y)) - 1 )
    """
    if y1 is None:
        x, y1, y_p1, y_p_var1 = get_processed_ifr(t1, events, b=binsize, window=window, zscore=False, convolve=0)
        y1 = (y1*binsize/1000).astype(np.int64)
    if y2 is None:
        x, y2, y_p2, y_p_var2 = get_processed_ifr(t2, events, b=binsize, window=window, zscore=False, convolve=0)
        y2 = (y2*binsize/1000).astype(np.int64)

    trial_ids = np.arange(y1.shape[0])
    cell_1_active_m = (np.mean(y1, axis=1) > 0)
    cell_2_active_m = (np.mean(y2, axis=1) > 0)
    cells_active_m = cell_1_active_m & cell_2_active_m
    if cells_active_m.sum()<20:
        print("WARNING two neurons share less than 20 trials in common. Cannot estimate synchrony.")
        if return_trials:
            y1.mean(axis=0)*np.nan, y1*np.nan, trial_ids[cells_active_m]
        return y1.mean(axis=0)*np.nan
    if remove_empty_trials:
        y1 = y1[cells_active_m, :]
        y2 = y2[cells_active_m, :]
        trial_ids = trial_ids[cells_active_m]

    if pre_smooth:
        if continuous:
            y1 = smooth(y1, 'gaussian_causal', smooth_sd)
            y2 = smooth(y2, 'gaussian_causal', smooth_sd)
        else:
            print('WARNING no pre-smoothing as arrays must be binary. Set continuous=True if you wish to pre smooth.')

    if continuous:
        # Max = 1
        y1[y1>1]=1
        y2[y2>1]=1

        # P(A,B)
        p_12_trials = y1*y2
        p_12 = np.mean(p_12_trials, 0)

        # P(A), P(B)
        p1 = np.mean(y1, 0)
        p2 = np.mean(y2, 0)

    else:
        # Binarize spike presence
        y1 = y1.astype(bool)
        y2 = y2.astype(bool)

        # P(A,B)
        p_12_trials = (y1 & y2).astype(int)
        p_12 = np.mean(p_12_trials, 0)

        # P(A), P(B)
        p1 = np.mean(y1.astype(int), 0)
        p2 = np.mean(y2.astype(int), 0)

    # P(A,B) / (P(A) x P(B))
    sync = p_12/(p1*p2)

    if post_smooth:
        sync = smooth(sync, 'gaussian_causal', smooth_sd)

    sync[np.isnan(sync)]=1

    if return_trials:
        if return_terms:
            ret = (sync, p_12_trials/(p1*p2), p_12_trials, p1, p2, trial_ids)
        else:
            ret = (sync, p_12_trials/(p1*p2), trial_ids)
    else:
        if return_terms:
            ret = (sync, p_12_trials, p1, p2)
        else:
            ret = sync

    return ret
    

def compute_sync_matrix(signal):
    """Compute the P(A int B)/P(A)P(B) for each neuron A and B,
    output is a N x N matrix.

    Works for continuous values [0-1] (does compute A&B, rather A*B: 11->1, 01->0, but .5 1 -> .5)
    """
    T, N = signal.shape
    neuron_fire_rate = signal.mean(axis=0)
    normalised_signal = signal / neuron_fire_rate
    sync_matrix_with_diag = normalised_signal.T.dot(normalised_signal) / T
    # simply get rid of diagonal (neurons with themselves)
    sync_matrix = sync_matrix_with_diag - np.diag(np.diag(sync_matrix_with_diag))
    return sync_matrix

def l2_synchrony(signal):
    T, N = signal.shape
    sync_matrix = compute_sync_matrix(signal)
    return np.sqrt((sync_matrix**2).sum()/(N*(N-1)))

def avg_synchrony(signal):
    T, N = signal.shape
    sync_matrix = compute_sync_matrix(signal)
    return sync_matrix.sum()/(N*(N-1))

# def two_synchrony(signal):
#     nb_neurons_active = signal.sum(axis=-1)
#     two_neurons_active = nb_neurons_active > 2
#     sync_gross = two_neurons_active.mean(axis=-1)
#     neuron_rate = signal.mean(axis=-2)
#     my_binary = int_to_binary(array, N)

def total_synchrony(signal, normalize=False, return_trials=False):
    """
    Signal: B x T x N tensor
    Computes ratio of N_neurons_active/N_neurons_active_if_were_independant.

    close to P Latham suggestion (sum across neurons).

    Limited between 0 and 1 if normalized, 0 and N if not.
    """
    N = signal.shape[-1]
    # Z = sum_i(Xi) with i=[1-N] neurons: that's the number of activated neurons = synchrony
    # total synchrony = ratio between Z**2 and Z**2 if neurons were independant
    # (we use Z**2 because it exaggerates differences in number of activated neurons)

    # CalcuLate E(Z**2)
    #     Var(Z) = E(Z**2) - (E(Z))**2 (definition of variance)
    # <=> E(Z**2) = Var(Z) + (E(Z))**2
    #             = Var(sum_i(Xi)) + E(sum_i(Xi))**2
    Z2_E = (signal.sum(axis=-1)**2).mean(axis=-1)
    # in theory same as Z2 = signal.sum(axis=-1).var(axis=-1) + signal.mean(axis=-2).sum(axis=-1)**2
    
    # If Xi are independant, Var(sum_i(Xi)) = sum_i(Var(Xi))
    # so E(Z**2) = Var(sum_i(Xi)) + E(sum_i(Xi))**2
    #            = sum_i(Var(Xi)) + E(sum_i(Xi))**2
    Z2_E_indep = signal.var(axis=-2).sum(axis=-1) + signal.mean(axis=-2).sum(axis=-1)**2

    # E(Z**2)/E_indep(Z**2)
    tot_sync = Z2_E/Z2_E_indep
    if normalize: tot_sync /= N

    if return_trials:
        # directly calculate Z**2, not E(Z**2) across trials
        Z2_trials = (signal.sum(axis=-1)**2)
        tot_sync_trials = Z2_trials/Z2_E_indep
        if normalize: tot_sync_trials /= N
        return tot_sync, tot_sync_trials

    return tot_sync

def total_var_synchrony(signal, normalize=True):
    """
    Signal: B x T x N tensor
    Same approach as total_synchrony (ratio between 2 value with numerator = true data and denominator = indep data),
    but using variances instead of means of squared data.
    Limited between 0 and 1 if normalized, 0 and N if not.

    Source: www.scholarpedia.org/article/Neuronal_synchrony_measures
    (written by David Golomb, citable publication is Golomb abd Rinzel 1993, Coherence metric; Ginzburg and Sompolinsky 1994.)
    but replace Vi(t) voltage of neuron i at time t by Pi(b) probability of spike of neuron i in bin b
    """

    # Z = sum_i(Xi) with i=[1-N] neurons: that's the number of activated neurons = synchrony

    Z_V = signal.sum(axis=-1).var(axis=-1)
    # same as Z_V = (signal.sum(axis=-1)**2).mean(axis=-1) - signal.sum(axis=-1).mean(axis=-1)**2
    
    # If Xi are independant, Var(sum_i(Xi)) = sum_i(Var(Xi))
    # so E(Z**2) = Var(sum_i(Xi)) + E(sum_i(Xi))**2
    #            = sum_i(Var(Xi)) + E(sum_i(Xi))**2
    Z_V_indep = signal.var(axis=-2).sum(axis=-1)

    # E(Z**2)/E_indep(Z**2)
    tot_var_sync = Z_V/Z_V_indep

    tot_var_sync = Z_V/Z_V_indep
    if normalize:
        N = signal.shape[-1]
        tot_var_sync /= N

    return np.sqrt(tot_var_sync)

def mgf_synchrony(signal, lam=2, normalized=True):
    """
    Signal: B x T x N matrix
    Computes moment generating function synchrony across N neurons.
    """
    assert lam>1, "WARNING lambda must be strictly superior to 1."
    # variance across trials (T) of the sum across neurons (N)
    mgf_synchrony = (lam**(signal.sum(axis=-1))).mean(axis=-1)
    if normalized:
        mgf_synchrony /= (lam**signal).mean(axis=-2).prod(axis=-1)

    return mgf_synchrony


### crosscorrelation between neuron population activation
### and target neuron

def lagged_synchrony_analysis(signal, target, lags, res=1):
    synchronic_signal = more_than_n_neurons_active(signal, res)
    return lagged_correlations(synchronic_signal, target, lags)

def more_than_n_neurons_active(signal, res=1):
    nb_neurons = signal.shape[-1]
    nb_active = signal.sum(axis=-1)
    L = []
    for n in np.arange(0,nb_neurons,res):
        L.append(nb_active > n)
    return np.stack(L, axis=-1)

def lagged_correlations(signal, target, lags, axis=1):
    L = [lagged_correlation(signal, target, lag, axis=axis) for lag in lags]
    return np.stack(L, axis=1)

def lagged_correlation(signal, target, lag, axis=1):
    target = np.expand_dims(target, axis=-1)
    target_lagged = np.roll(target, -lag, axis=1)
    cor = correlation(signal, target_lagged, axis=axis)
    return cor

def correlation(x, y, axis=1):
    cov = (x * y).mean(axis=axis) - x.mean(axis=axis) * y.mean(axis=axis)
    sdx = x.std(axis=axis)
    #sdx = np.nan_to_num(sdx, nan=1)
    sdy = y.std(axis=axis)
    #sdy =np.nan_to_num(sdy, nan=1)
    cor = cov / (sdx * sdy)

    return cor



### mutual information between population of PCs and NC
def multivariate_mutual_information(signal_x, signal_y):
    """
    signal_x is a B x T x N matrix
    signal_y is a B x T matrix
    """
    p_joint_x = compute_p_joint(signal_x).mean(axis=-1).T
    H_joint_x = entropy(p_joint_x)

    p_y = signal_y.mean(axis=-1)
    H_y = - p_y * cut_log(p_y) - (1-p_y) * cut_log(1-p_y)

    signal_xy = np.concatenate([signal_x, signal_y[:,:,None]], axis=-1)
    p_joint_xy = compute_p_joint(signal_xy).mean(axis=-1).T
    H_joint_xy = entropy(p_joint_xy)

    return (H_joint_x + H_y - H_joint_xy)/(H_joint_x + H_y)


### Total correlation i.e. generalization of mutual information
### to find consistent patterns of codependance across N neurons

def total_correlation(signal, normalized=True):
    """
    Generalization to N variables of mutual information (for 2 variables, same thing).

    i.e. KL divergence from the joint distribution P(X1,...,Xn)
    to the product (=independant) distribution P(X1)*...*P(Xn).

    If there are only 2 variables, this corresponds to the mutual information.

    This reduces to the simpler difference of entropies,
    sum_over_i(H(Xi)) - H(X1,...,Xn)
    
    Returns the total_correlation normalised between 0 and 1
    for a B x T x N or T x N signal of probabilities between 0 and 1.
    
    If normalize=True, normalized between 0 and 1."""
    # get joint entropy
    # that is across neurons
    p_joint = compute_p_joint(signal)
    p_joint_mean = p_joint.mean(axis=-1).T
    joint_entropy = entropy(p_joint_mean)

    neuron_fire_rate = signal.mean(axis=-2)
    neuron_entropies = - neuron_fire_rate * cut_log(neuron_fire_rate) - (1-neuron_fire_rate) * cut_log(1-neuron_fire_rate)

    # this is equivalent to kullback_leibler(p_joint, p_product)
    C = neuron_entropies.sum(axis=-1) - joint_entropy

    if normalized:
        C_max = neuron_entropies.sum(axis=-1) - neuron_entropies.max(axis=-1)
        C /= C_max
    
    return C

def mutual_information(signal):
    """KL divergence from the joint distribution P(X1,...,Xn)
    to the product (=independant) distribution P(X1)*...*P(Xn)
    
    returns the mutual information for a B x T x N or T x N signal of probabilities between 0 and 1"""
    p_joint = compute_p_joint(signal)
    p_prod = compute_p_prod(signal)
    KL = kullback_leibler(p_joint.mean(axis=-1).T, p_prod.mean(axis=-1).T)
    return KL

def multivariate_copula(signal, normalized=True):
    """Compute the proba Q such that P(X1, ..., XN) ~= Q(X1, ..., XN)P(X1)...P(XN).
    
    If normalize=True,
    return H(Q)/H(Uniform) with Q the copula of the signal
    (divergence between Q and the uniform distribution).
    Value is 0 if independant, 1 if all neurons always fire the same.
    """
    p_joint = compute_p_joint(signal).mean(axis=-1).T
    p_prod = compute_p_prod2(signal)
    q = p_joint / p_prod
    q = q / q.sum(axis=-1, keepdims=True)

    if normalized:
        # divergence between copula and the uniform distribution
        # i.e. ratio of their entropies
        N = signal.shape[-1]
        q = 1 - (entropy(q) / entropy(np.ones(2**N)/2**N))

    return q



### basic information theoretic functions ###

def compute_p_joint(signal):
    """returns the product proba of the configuration: p_joint(config),
    output an array of size 2^N.

    p_joint is of shape (B, 2^N)
    where 2^N corresponds to all combinations of neurons: 0000, 0001, 0010, 0011...
    so p_joint[:,-1] corresponds to the joint probability taking all neurons into consideration
    (P(X1 & X1 & ... & Xn))
    """

    # illustrative example:
    # np.all(p_joint[-1,:,:] == np.logical_and.reduce(M, axis=-1))

    binaries = array_of_all_binaries(signal.shape)
    p_joint = equivalence_measure(signal, binaries)
    return p_joint

def compute_p_prod(signal):
    """returns the proba of the configuration: p_product(config),
    output an array of size 2^N.
    """
    binaries = array_of_all_binaries(signal.shape)
    neuron_fire_rate = signal.mean(axis=-2, keepdims=True)
    p_prod = equivalence_measure(neuron_fire_rate, binaries)
    return p_prod

def compute_p_prod2(signal):
    """returns the proba of the configuration: p_joint(config),
    output an array of size (B,2^N)
    (2^N is the number of combinations of N neurons, for instance 32 for 5 neurons).
    """
    N = signal.shape[-1]
    integers = np.arange(2**N).reshape((2**N, 1))
    binaries = int_to_binary(integers, N)
    neuron_fire_rate = signal.mean(axis=-2, keepdims=True)
    p_prod = equivalence_measure(neuron_fire_rate, binaries)
    return p_prod

def kullback_leibler(p, q):
    """Compute the KL divergence of two probability measures, given by arrays of the same size that sum to one."""
    log_ratio = np.log(p / q, out=np.zeros(p.shape), where=(p != 0))
    KL = (p * log_ratio).sum(axis=-1)
    return KL

def entropy(p, axis=-1):
    return -(p * cut_log(p)).sum(axis=axis)



### utility functions ###

def array_of_all_binaries(signal_shape):
    """returns an array with all binaries in {0, 1}^N, 
    with shape (2^N, 1, ..., 1, N) with the same nb of ones as the dimensions of the signal tensor minus one.
    """
    N = signal_shape[-1]
    shape = broadcastable_shape(2**N, len(signal_shape))
    integers = np.arange(2**N).reshape(shape)
    binaries = int_to_binary(integers, N)
    return binaries

def broadcastable_shape(m, n):
    return (m, *(n*[1]))

def int_to_binary(array, N):
    """A dimension is added to a tensor of integers, such that the last dimension gives the binary decomposition,
    output a tensor with one more dimension of size N.
    """
    indexes = np.arange(N)
    res = (array // 2**indexes) % 2
    return res

def equivalence_measure(p, q):
    """Give the probability that the two measures on {0,1}^N are equal,
    with discrete binaries, it is equivalent to p == q,
    output an tensor of the size broadcasted from the two, minus the last.
    """
    return (p*q + (1-p)*(1-q)).prod(axis=-1)

def cut_log(p):
    """returns log p if p > 0, else 0."""
    return np.log(p, out=np.zeros(p.shape), where=(p > 0))


#####################

def residual_cv2(t, bin, win, events):
    '''

    Returns:
        - cv2_residuals: nevents x time matrix np array
    '''

    

    # Generate events matrix


    # Subtract mean from events matrix

    return

def Paintb_PaPb(y1, y2):
    """
    y1 and y2 should be T x B matrices (T trials and B bins)
    """

    # Binarize spike presence
    y1_bool = y1.astype(bool)
    y2_bool = y2.astype(bool)

    # P(A,B)
    p_12 = np.mean((y1_bool & y2_bool).astype(int), 0)

    # P(A), P(B)
    p1 = np.mean((y1_bool).astype(int), 0)
    p2 = np.mean((y2_bool).astype(int), 0)

    # P(A,B) / (P(A) x P(B))
    sync = p_12/(p1*p2)

    return sync


# -*- coding: utf-8 -*-
"""
2018-07-20
@author: Maxime Beau, Neural Computations Lab, University College London
"""
from IPython.core.debugger import set_trace as breakpoint 
import os.path as op
opj=op.join
from pathlib import Path, PosixPath, WindowsPath

# from itertools import groupby
# from operator import itemgetter

import matplotlib.pyplot as plt

import numpy as np
import pandas as pd
from scipy.stats import iqr
from scipy.optimize import curve_fit
from scipy.stats import norm
from npyx.utils import smooth, thresh_consec, npa, assert_int, assert_float
from npyx.gl import get_units, get_npyx_memory, check_periods
from npyx.inout import read_metadata


def ids(dp, unit, sav=True, verbose=False, periods='all', again=False):
    '''
    ********
    routine from routines_spikes
    computes spike indices (1, Nspikes) - int64, in samples
    ********

    - dp (string): DataPath to the Neuropixels dataset.
    - u (int): unit index
    - ret (bool - default False): if True, train returned by the routine.
      If False, by definition of the routine, drawn to global namespace.
    - sav (bool - default True): if True, by definition of the routine, saves the file in dp/routinesMemory.
    - periods = 'all' or [(t1, t2), (t3, t4), ...] with t1, t2 in seconds.
    - again: boolean, if True recomputes data from source files without checking routines memory.
    '''

    dp = Path(dp)
    assert unit in get_units(dp), 'WARNING unit {} not found in dataset {}!'.format(unit, dp)
    # Search if the variable is already saved in dp/routinesMemory
    dpnm = get_npyx_memory(dp)

    fn='ids{}.npy'.format(unit)
    if op.exists(Path(dpnm,fn)) and not again:
        if verbose: print("File {} found in routines memory.".format(fn))
        indices = np.asarray(np.load(Path(dpnm,fn)), dtype='int64')
    # if not, compute it
    else:
        if verbose: print(f"File {fn} not found in routines memory. Will be computed from source files.")
        if not (assert_int(unit)|assert_float(unit)): raise TypeError(f'WARNING unit {unit} type ({type(unit)}) not handled!')
        assert unit in get_units(dp), f'WARNING unit {unit} not found in dataset {dp}!'
        # if assert_multi(dp):
        #     ds_table = get_ds_table(dp)
        #     if ds_table.shape[0]>1: # If merged dataset
        #         spike_clusters = np.load(Path(dp,"spike_clusters.npy"), mmap_mode='r')
        #         indices = np.nonzero(spike_clusters==unit)[0].ravel()
        #     else:
        #         ds_i, unt = get_dataset_id(unit)
        #         spike_clusters = np.load(Path(ds_table.loc['dp'][ds_i],"spike_clusters.npy"), mmap_mode='r')
        #         indices = np.nonzero(spike_clusters==unt)[0].ravel()
        # else:

        spike_clusters = np.load(dp/"spike_clusters.npy", mmap_mode='r')
        indices = np.nonzero(spike_clusters==unit)[0].ravel()

        # Save it
        if sav:
            np.save(dpnm/fn, indices)

    # Optional selection of a section of the recording.
    # Always computed because cannot reasonably be part of file name.
    periods = check_periods(periods)
    if periods is not 'all': # else, eq to periods=[(0, spike_samples[-1])] # in samples
        dp_source = get_source_dp_u(dp, unit)[0]
        fs=read_metadata(dp_source)["highpass"]['sampling_rate']
        train=trn(dp, unit, again=again)
        sec_bool=np.zeros(len(train), dtype=np.bool)
        for section in periods:
            sec_bool[(train>=section[0]*fs)&(train<=section[1]*fs)]=True # comparison in samples
        indices=indices[sec_bool]

    return indices


def trn(dp, unit, sav=True, verbose=False, periods='all', again=False, enforced_rp=0):
    '''
    ********
    routine from routines_spikes
    computes spike train (1, Nspikes) - int64, in samples
    ********

    - dp (string): DataPath to the Neuropixels dataset.
    - u (int): unit index
    - ret (bool - default False): if True, train returned by the routine.
    if False, by definition of the routine, drawn to global namespace.
    - sav (bool - default True): if True, by definition of the routine, saves the file in dp/routinesMemory.
    - periods: list [[t1,t2], [t3,t4],...] (in seconds) or 'all' for all periods.
    - again: boolean, if True recomputes data from source files without checking routines memory.
    - enforced_rp: enforced refractory period, in millisecond. 0 by default (only pure duplicates are removed)
    '''

    # Search if the variable is already saved in dp/routinesMemory
    dpnm = get_npyx_memory(dp)
    dp_source = get_source_dp_u(dp, unit)[0]
    fs=read_metadata(dp_source)['highpass']['sampling_rate']

    fn=f'trn{unit}_{enforced_rp}.npy'
    if (dpnm/fn).exists() and not again:
        if verbose: print("File {} found in routines memory.".format(fn))
        try: train = np.load(dpnm/fn) # handling of weird allow_picke=True error
        except: pass

    # if not, compute it
    if 'train' not in locals(): # handling of weird allow_picke=True error when using joblib multiprocessing
        if verbose: print(f"File {fn} not found in routines memory. Will be computed from source files.")
        if not (assert_int(unit)|assert_float(unit)): raise TypeError(f'WARNING unit {unit} type ({type(unit)}) not handled!')
        assert unit in get_units(dp), f'WARNING unit {unit} not found in dataset {dp}!'
        # if assert_multi(dp):
        #     ds_table = get_ds_table(dp)
        #     if ds_table.shape[0]>1: # If merged dataset
        #         spike_clusters = np.load(Path(dp,"spike_clusters.npy"), mmap_mode='r')
        #         spike_samples = np.load(Path(dp,'spike_times.npy'), mmap_mode='r')
        #         train = spike_samples[spike_clusters==unit].ravel()
        #     else:
        #         ds_i, unt = get_dataset_id(unit)
        #         spike_clusters = np.load(Path(ds_table['dp'][ds_i],"spike_clusters.npy"), mmap_mode='r')
        #         spike_samples = np.load(Path(ds_table['dp'][ds_i],'spike_times.npy'), mmap_mode='r')
        #         train = spike_samples[spike_clusters==unt].ravel()
        # else:
        spike_clusters = np.load(Path(dp,"spike_clusters.npy"), mmap_mode='r')
        spike_samples = np.load(Path(dp,'spike_times.npy'), mmap_mode='r')
        train = spike_samples[spike_clusters==unit].ravel()

        # Filter out spike duplicates (spikes following an ISI shorter than enforced_rp)
        # by default, only pure duplicates (yeah they happen!!)
        assert len(train)!=0, f'unit {unit} not found in spike_clusters.npy - probably a merger bug.'
        train=train[np.append(True, np.diff(train)>=enforced_rp*fs/1000)]

        # Save it
        if sav:
            np.save(dpnm/fn, train)

    # Optional selection of a section of the recording.
    # Always computed because cannot reasonably be part of file name.
    periods = check_periods(periods)
    if periods is not 'all': # else, eq to periods=[(0, spike_samples[-1])] (in samples)
        sec_bool=np.zeros(len(train), dtype=np.bool)
        for section in periods:
            sec_bool[(train>=section[0]*fs)&(train<=section[1]*fs)]=True # comparison in samples
        train=train[sec_bool]

    return train

def isi(dp, unit, enforced_rp=0, sav=True, verbose=False, periods='all', again=False):
    '''
    ********
    routine from routines_spikes
    computes spike inter spike intervals (1, Nspikes-1) - float64, in milliseconds
    ********

    - dp (string): DataPath to the Neuropixels dataset.
    - u (int): unit index
    - ret (bool - default False): if True, train returned by the routine.
      If False, by definition of the routine, drawn to global namespace.
    - sav (bool - default True): if True, by definition of the routine, saves the file in dp/routinesMemory.
    '''
    t=trn(dp, unit, sav, verbose, periods, again, enforced_rp)
    return np.diff(t) if len(t)>1 else None

def inst_cv2(t):
    '''
    Parameters:
        - t: (nspikes,) np array, spike times in any unit

    Returns:
        - cv2: (nspikes-2,) array, instantaneous cv2
    '''

    isint = np.diff(t)

    cv2 = 2 * np.abs(isint[1:] - isint[:-1]) / (isint[1:] + isint[:-1])

    return cv2

def mean_firing_rate(t, exclusion_quantile=0.005, fs=30000):
    i = np.diff(t) if len(t)>1 else None
    if i is None: return 0
    # Remove outlyers
    i=i[(i>=np.quantile(i, exclusion_quantile))&(i<=np.quantile(i, 1-exclusion_quantile))]/fs
    return np.round(1./np.mean(i),2)

def mfr(dp=None, U=None, exclusion_quantile=0.005, enforced_rp=0,
        periods='all', again=False, train=None, fs=None):
    '''
    Computes the mean firing rate of a unit.

    2 ways to use it:
     - regular npyx mfr(dp, u) syntax - will grab the spike times and fs from files
     - custom mfr(train=array, fs=sampling_frequency) - for any spike train

    Parameters:
        - dp: str, datapath
        - U: int or list, unit or list of units
        - exclusion_quantile: float, quantiles beyond which we exclude interspike intervals (very short or very long)
        - enforced_rp: float, enforced refractory period - minimum time in ms separating consecutive spikes of the train - if sorther than that, the 1st spike of the couple is kept.
        - periods: 'all' or [(t1,t2), ...] - periods in s to consider for calculation
        - again=False
        - train=None
        - fs=None

    Returns:
        - mfr: float or np array, mean firing rates of unit(s) in Hz
    '''
    if train is not None:
        assert fs is not None, 'you need to provide a sampling frequency!'
        train=np.asarray(train)
        assert train.ndim==1
        return mean_firing_rate(train, exclusion_quantile, fs)

    U=npa([U]).flatten()
    MFR=[]
    for u in U:
        t=trn(dp, u, periods=periods, again=again, enforced_rp=enforced_rp)
        dp_source = get_source_dp_u(dp, u)[0]
        fs=read_metadata(dp_source)['highpass']['sampling_rate']
        MFR.append(mean_firing_rate(t, exclusion_quantile, fs))

    return MFR[0] if len(U)==1 else npa(MFR)

def binarize(X, bin_size, fs, rec_len=None):
    '''Function to turn a spike train (array of time stamps)
       into a binarized spike train (array of 0 or 1
                                     of length rec_len with a given bin_size.).
       - X: spike train (array of time stamps, in samples sampled at fs Hertz)
       - bin_size: size of binarized spike train bins, in milliseconds.
       - rec_len: length of the recording, in SAMPLES. If not provided, time of the last spike.
       - fs: sampling frequency, in Hertz.'''

    # Process bin_size
    bin_size = int(np.ceil(fs*bin_size/1000))  # Conversion ms->samples

    # Process rec_len
    if rec_len is None:
        rec_len=X[-1]

    # Binarize spike train
    Xb = np.histogram(X, bins=np.arange(0, rec_len, bin_size))[0]

    # Decrease array size as much as possible
    for encode in [32,16,8]:
        Xb1=Xb.astype(f'int{encode}')
        if not np.all(Xb==Xb1):
            break
        Xb=Xb1

    return Xb

def trnb(dp, u, b, periods='all', again=False):
    '''
    ********
    routine from routines_spikes
    computes binarized spike train (1, Nspikes) - int64, in samples
    ********

    - dp (string): DataPath to the Neuropixels dataset.
    - u (int): unit index
    - bin_size: size of binarized spike train bins, in milliseconds.
    '''
    dp_source = get_source_dp_u(dp, u)[0]
    fs=read_metadata(dp_source)['highpass']['sampling_rate']
    assert b>=1000/fs
    t = trn(dp, u, enforced_rp=1, periods=periods, again=again)
    t_end = np.load(Path(dp,'spike_times.npy'), mmap_mode='r').ravel()[-1]
    return binarize(t, b, fs, t_end)

def get_firing_periods(dp, u, b=1, sd=1000, th=0.02, again=False, train=None, fs=None, t_end=None):
    '''
    Parameters:
        - t: array of spike times, in samples
        - t_end: recording end time, in samples
        - b: float, bin size i.e. temporal resolution of presence periods, in ms | Default 1
        - sd: float, standard deviation of gaussian smoothing window, in ms | Default 1000
        - th: threshold to define presence, in fraction of mean firing rate
        - fs: sampling rate of spike times, in Hz
    '''
    assert 1<sd<10000
    assert 0<=th<1
    sav=False
    if train is None:
        sav=True
        dpnm = get_npyx_memory(dp)
        fn=f'firing_periods_{u}_{b}_{sd}_{th}.npy'
        if op.exists(Path(dpnm,fn)) and not again:
            return np.load(Path(dpnm,fn))
        t = trn(dp, u, enforced_rp=1, again=again)
        dp_source = get_source_dp_u(dp, u)[0]
        fs=read_metadata(dp_source)['highpass']['sampling_rate']
        t_end = np.load(Path(dp,'spike_times.npy'), mmap_mode='r').ravel()[-1]
    else:
        assert fs is not None, "You need to provide the sampling rate of the provided train!"
        assert t_end is not None, "You need to provide the end time 't_end' of recorded train that you provide!"
        t=np.asarray(train)
        assert t.ndim==1

    periods = firing_periods(t, fs, t_end, b=1, sd=1000, th=0.02)

    if sav:
        np.save(Path(dpnm,fn), periods)

    return periods

def firing_periods(t, fs, t_end, b=1, sd=1000, th=0.02, again=False, dp=None, u=None):
    '''
    Parameters:
        - t: array of spike times, in samples
        - fs: sampling rate of spike times, in Hz
        - t_end: recording end time, in samples
        - b: float, bin size i.e. temporal resolution of presence periods, in ms | Default 1
        - sd: float, standard deviation of gaussian smoothing window, in ms | Default 1000
        - th: threshold to define presence, in fraction of mean firing rate
    '''
    sav=False
    if u is not None:
        sav=True
        assert dp is not None
        assert len(trn(dp,u,0))==len(t), 'There seems to be a mismatch between the provided spike trains and the unit index.'
        fn=f'firing_periods_{u}_{b}_{sd}_{th}.npy'
        dpnm = get_npyx_memory(dp)
        if op.exists(Path(dpnm,fn)) and not again:
            return np.load(Path(dpnm,fn))

    assert 1<sd<10000
    assert 0<=th<1
    assert t.ndim==1
    t=np.asarray(t)

    assert b>=1000/fs
    tb = binarize(t, b, fs, t_end)
    sd=int(sd/b) # convert from ms to bin units
    b_s=b/1000 # bin seconds
    tbs=smooth(tb, 'gaussian', sd=sd)/b_s # result is inst. firing rate in Hz - speed bottleneck
    fr_th=mean_firing_rate(t, 0.005, fs)*th

    periods = thresh_consec(tbs, fr_th, sgn=1, n_consec=0, exclude_edges=False, only_max=False, ret_values=False)
    if not any(periods): periods=[[0,len(tbs)-1]]
    periods=(np.array(periods)*(b_s*fs)).astype(np.int64) # conversion from bins to samples

    if sav: np.save(Path(dpnm,fn), periods)

    return periods


def train_quality(dp, unit, period_m=[0,20],
                  fp_chunk_span=3, fp_chunk_size = 10,
                  fn_chunk_span = 3, fn_chunk_size = 10,
                  use_or_operator = True,
                  violations_ms = 0.8, fp_threshold = 0.05, fn_threshold = 0.05,
                  again = False, save = True, verbose = False, plot_debug = False):

    """
    Subselect spike times which meet two criteria:
        low number of 'missed spikes' (false negatives)
        and low number of 'extra spikes' (false positives).
        
    The recording within period_m is split in fp/fn_chunk_size seconds chunks,
    and fp/fn_chunk_span chunks are used to estimate the fp and fn rates.
    (e.g. 3 10s chunks means that the rates are estimated on 30s chunks, overlapping by 10s).

    - False negative rate estimation: for checking which sections of the recording
        have too many spikes missing, by looking if the section is
        approximately Gaussian. If the time section of the recording
        has too much of the Gaussian distribution cut off (>5%) the section
        has to be discarded
    - False positive rate estimation: check if there are not too many spikes
        occuring in the the refractory period of the autocorrelogram.
        If there are too many spikes in the ACG the section of the
        recording will be discarded.
        
    Finally, the spikes belonging to the intersection (use_or_operator=False) or union (use_or_operator=True)
    of the chunks with low enough fp/fn rates are returned.
    
    E.g. if fp_chunk_span=3 and fp_chunk_size=10 anf use_or_operator=False,
    for a given 10s "chunk_k" to be considered, ALL 3*10=30s chunk triplets
    (chunk_k-2,chunk_k-1,chunk_k), (chunk_k-1,chunk_k,chunk_k+1) AND (chunk_k,chunk_k+1,chunk_k+2)
    fp rate estimations must be below fp_threshold.
    
    ***********
    
    Parameters:
        - dp: str, path to dataset.
        - unit: int/float, unit index.
        - period_m: [t1, t2] list of floats in minutes, period to consider
        - fp_chunk_span: int, number of recording chunks to concatenate to estimate fp rate.
        - fp_chunk_size: int, size of recording chunks used to estimate fp rate.
        
        - fn_chunk_span: int, number of chunks to concatenate to estimate fp rate.
        - fn_chunk_size: int, size of recording chunks used to estimate fn rate.
        
        - use_or_operator: bool, whether to use the union (True) or intersection (False)
                        of stitched chunks (fp/fn_chunk_span)
                        to state that a chunk passed the fp/fn threshold or not.
                        E.g. if 
        
        - violations_ms: float, width of window in ms used to estimate refractory period violations (center of autocorrelogram)
        - fp_threshold: [0-1] float, false positive rate (ratio of refractory periods violations/mean firing rate) threshold
        - fn_threshold: [0-1] float, false negative rate (AUC of gaussian fit missing) threshold
        
        - again: bool, whether to recompute trn_filtered or simply load it from npyxMemory
        - save: bool, whether to save result to npyxMemory for future fast reloading
        - verbose: bool, whether to print extra information for debugging purposes.
    
    Returns:
    - goodsec, acgsec, gausssec

    """

    # Hard-coded parameters
    c_bin = 0.2
    c_win = 100
    n_bins_acg_baseline=80 # from start and end of acg window
    n_spikes_threshold = 300
    fs = 30_000


    # check that the passed values make sense
    assert isinstance(dp, (str, PosixPath, WindowsPath)),\
        'Provide a string or a pathlib object as the source directory'
    dp = Path(dp)

    assert assert_int(unit), 'Unit provided should be an int'
    assert assert_int(fp_chunk_span), 'fp_chunk_span provided should be an int'
    assert assert_int(fp_chunk_size), 'fp_chunk_size provided should be an int'
    assert assert_int(fn_chunk_span), 'fn_chunk_span provided should be an int'
    assert assert_int(fn_chunk_size), 'fn_chunk_size provided should be an int'

    assert fp_chunk_size >= 1, "ACG window length needs to be larger than 1 sec"
    assert fn_chunk_size >= 1, "Gaussian window length needs to be larger than 1 sec"
    assert fp_chunk_span >= 1, "ACG chunk size needs to be larger than 1 "
    assert fn_chunk_span >= 1, "Gaussian chunk size needs to be larger than 1 "

    dpnm = get_npyx_memory(dp)
    
    # Load data
    unit_ids = ids(dp, unit, again=again, verbose=verbose)
    unit_amp = np.load(dp/'amplitudes.npy')[unit_ids]
    
    unit_train = trn(dp, unit, enforced_rp=0, again=again, verbose=verbose)/fs
    period_s=[period_m[0]*60, period_m[1]*60]

    # Attempt to reload if precomputed
    fn=(f"trn_qual_{unit}_{str(period_m).replace(' ','')}"
        f"_{str(fp_chunk_span)}_{str(fp_chunk_size)}_{str(fn_chunk_span)}_{str(fn_chunk_size)}"
        f"_{str(violations_ms)}_{str(fp_threshold)}_{str(fn_threshold)}.npy")
    fn_spikes = "spikes_"+fn
    if (dpnm/fn).exists() and (dpnm/fn_spikes).exists() and (not again):
        if verbose: print(f"File {fn} found in routines memory.")
        good_fp_start_end, good_fn_start_end = np.load(dpnm/fn, allow_pickle=True)
        good_spikes_m = np.load(dpnm/fn_spikes)
        
        good_fp_start_end_plot=None if len(good_fp_start_end)==1 else good_fp_start_end
        good_fn_start_end_plot=None if len(good_fn_start_end)==1 else good_fn_start_end
        if plot_debug:
            plot_fp_fn_rates(unit_train, period_s, unit_amp, good_spikes_m,
                None, None, None, None,
                fp_threshold, fn_threshold, good_fp_start_end_plot, good_fn_start_end_plot)
        
        return good_spikes_m, good_fp_start_end.tolist(), good_fn_start_end.tolist()
    
    n_spikes = np.count_nonzero((unit_train>period_s[0])&(unit_train<period_s[1]))

    # steps:
        # split into 10 second chunks
        # run 3 concatenated 10 sec chunks through the filter
        # find all the chunks that passed the filter
        # find the time winodws where thes chunks happened
        # extract the spike times again from these consecutive windows
        # append the spike times from these consecutiv windowws
        # calculate features on this array

    recording_span = period_s[1]-period_s[0]
    n_fn_chunks =  int(recording_span / fn_chunk_size)
    n_fp_chunks =  int(recording_span / fp_chunk_size)
    passed_fn = np.zeros((n_fn_chunks,3)).astype('int')
    passed_fp = np.zeros((n_fp_chunks,3)).astype('int')
    fn_chunks = [[period_s[0]+t*fn_chunk_size, period_s[0]+(t+fn_chunk_span)*fn_chunk_size] for t in range(n_fn_chunks)]
    fp_chunks = [[period_s[0]+t*fp_chunk_size, period_s[0]+(t+fp_chunk_span)*fp_chunk_size] for t in range(n_fn_chunks)]


    fp_toplot, chunk_fp_t, fn_toplot, chunk_fn_t = [], [], [], []
    if len(unit_ids) > n_spikes_threshold:
        
        # False negative estimation
        for i, (t1,t2) in enumerate(fn_chunks):
            chunk_mask = (t1 <= unit_train) & (unit_train < t2)
            n_spikes_chunk=np.sum(chunk_mask)

            if n_spikes_chunk > 15:
                amplitudes_chunk = unit_amp[chunk_mask].astype(np.float64)
                chunk_bins = estimate_bins(amplitudes_chunk, rule='Fd')
                if chunk_bins> 3:
                    x_c, p0_c, min_amp_c, n_fit_c, n_fit_no_cut_c, chunk_spikes_missing = gaussian_amp_est(amplitudes_chunk, chunk_bins)
                    fn_toplot.append(chunk_spikes_missing/100)
                    chunk_fn_t.append(t1+(t2-t1)/2)
                    if (~np.isnan(chunk_spikes_missing)) & (chunk_spikes_missing <= fn_threshold*100):
                        passed_fn[i] = [t1, t2, 1]

        # next loop through the chunks made for the ACG extraction
        # get the periods where the ACG filter passed
        
        # Compute denominator of FP on total period
        acg_tot = acg(dp, unit, c_bin, c_win, verbose = False,  periods=[period_s])
        x_block = np.round(np.arange(-c_win/2, c_win/2 + c_bin, c_bin), 8) # round to fix binary imprecisions
        rp_mask = (x_block >= -violations_ms) & (x_block <= violations_ms)
        baseline_mask = (acg_tot*0).astype(bool)
        baseline_mask[:n_bins_acg_baseline] = True
        baseline_mask[-n_bins_acg_baseline:] = True
        baseline_mean = np.mean(acg_tot[baseline_mask])
        
        # False positive estimation
        for i, (t1,t2) in enumerate(fp_chunks):

            chunk_mask = (t1 <= unit_train) & (unit_train < t2)
            n_spikes_chunk=np.sum(chunk_mask)

            if n_spikes_chunk > 15:
                ACG = acg(dp, unit, c_bin, c_win, verbose = False,  periods=[(t1, t2)])
                violations_mean = np.mean(ACG[rp_mask])
                rpv_ratio_acg = round(violations_mean / baseline_mean, 4)
                fp_toplot.append(rpv_ratio_acg)
                chunk_fp_t.append(t1+(t2-t1)/2)
                if (rpv_ratio_acg <= fp_threshold):
                    passed_fp[i] = [t1, t2, 1]

    # Across all chunks, if at least 1 good chunk for both (else no spike can be called good)
    if (np.sum(passed_fp[:,2])  > 1) & (np.sum(passed_fn[:,2]) > 1):

        # Aggregate FP and FN masks for each elemental chunks
        # e.g. for 30s chunks overlapping at 33% (3 10s chunks),
        # find for each 10s chunk what the final FP/FN rate is
        # based on the rate of the 3 30s chunks which overlap with it (and/or, see use_or_operator)
        # (apart from edges, where only 1 30s chunk overlaps)
        subchunks_fp = np.unique(npa(fp_chunks).flatten())
        subchunks_fp = npa([[subchunks_fp[i], subchunks_fp[i+1]] for i in range(len(subchunks_fp)-1)])
        good_fp_bool = passed_fp[:,2].astype(bool)
        subchunks_fp_bool = np.zeros((subchunks_fp.shape[0], fp_chunk_span))*np.nan
        for i in range(fp_chunk_span):
            subchunks_fp_bool[i:len(subchunks_fp_bool)-fp_chunk_span+i+1,i]=good_fp_bool
        
        subchunks_fn = np.unique(npa(fn_chunks).flatten())
        subchunks_fn = npa([[subchunks_fn[i], subchunks_fn[i+1]] for i in range(len(subchunks_fn)-1)])
        good_fn_bool = passed_fn[:,2].astype(bool)
        subchunks_fn_bool = np.zeros((subchunks_fn.shape[0], fn_chunk_span))*np.nan
        for i in range(fn_chunk_span):
            subchunks_fn_bool[i:len(subchunks_fn_bool)-fn_chunk_span+i+1,i]=good_fn_bool
        
        # Find out if subchunks are good based on overlapping chunks 
        # (NaNs do not count in np/all/any -> decision on edges based on only 1 chunk)
        nanm_fp = np.isnan(subchunks_fp_bool)
        nanm_fn = np.isnan(subchunks_fn_bool)
        if use_or_operator:
            # nans must be 0 to be ignored by any
            subchunks_fn_bool[nanm_fn] = 0
            subchunks_fp_bool[nanm_fp] = 0
            
            subchunks_fn_bool = np.any(subchunks_fn_bool, axis=1)
            subchunks_fp_bool = np.any(subchunks_fp_bool, axis=1)
        else:
            # nans must be 1 to be ignored by all
            subchunks_fn_bool[nanm_fn] = 1
            subchunks_fp_bool[nanm_fp] = 1
            
            subchunks_fn_bool = np.all(subchunks_fn_bool, axis=1)
            subchunks_fp_bool = np.all(subchunks_fp_bool, axis=1)
            
        good_fp_start_end = subchunks_fp[subchunks_fp_bool]
        good_fn_start_end = subchunks_fn[subchunks_fn_bool]
        
        ## Finally, mask spikes meeting the FP AND FN rates
        fp_m = np.zeros(unit_train.shape[0]).astype(bool)
        for subchunk in good_fp_start_end:
            m = (unit_train>subchunk[0])&(unit_train<subchunk[1])
            fp_m = fp_m|m
        fn_m = np.zeros(unit_train.shape[0]).astype(bool)
        for subchunk in good_fn_start_end:
            m = (unit_train>subchunk[0])&(unit_train<subchunk[1])
            fn_m = fn_m|m
        good_spikes_m=fp_m&fn_m

        if save:
            np.save(Path(dpnm,fn), np.array( (np.array(good_fp_start_end), np.array(good_fn_start_end)), dtype = object))
            np.save(dpnm/fn_spikes, good_spikes_m)

        if plot_debug:
            plot_fp_fn_rates(unit_train, period_s, unit_amp, good_spikes_m,
                     fp_toplot, fn_toplot, chunk_fp_t, chunk_fn_t,
                     fp_threshold, fn_threshold, good_fp_start_end, good_fn_start_end)

        return good_spikes_m, good_fp_start_end, good_fn_start_end
    
    else:
        good_spikes_m=(unit_train*0).astype(bool)
        if save:
            np.save(Path(dpnm,fn), np.array(([0], [0]), dtype = object)  )
            np.save(dpnm/fn_spikes, good_spikes_m)
            
        if plot_debug and n_spikes>0:
            plot_fp_fn_rates(unit_train, period_s, unit_amp, good_spikes_m,
                     fp_toplot, fn_toplot, chunk_fp_t, chunk_fn_t,
                     fp_threshold, fn_threshold)
            
        return good_spikes_m, [0], [0]

# def get_consec_sections(seconds):
#         """
#         Given an array with seconds as entries (with 1 sec increments)

#         Return: list of consecutive sections start and end times
#         """
#         sec_all = []
#         for k, g in groupby(enumerate(seconds), lambda ix: ix[0]-ix[1]):
#             sec_all.append(list( map(itemgetter(1), g)))
#         start_end = []

#         # get the start and end times of these section
#         for good_section in sec_all:
#             start_end.append([good_section[0], good_section[-1]])

#         return start_end


def trn_filtered(dp, unit, period_m=[0,20],
                  fp_chunk_span=3, fp_chunk_size = 10,
                  fn_chunk_span = 3, fn_chunk_size = 10,
                  use_or_operator = True,
                  violations_ms = 0.8, fp_threshold = 0.05, fn_threshold=0.05,
                  use_consecutive = False, consecutive_n_seconds = 180,
                  again = False, save = True, verbose = False, plot_debug = False):
    f"""
    Returns spike times (in sample) meeting the false positive and false negative criteria.
    Mainly wrapper of train_quality().
    
    Extra parameters not in train_quality (see below for others):
        - use_consecutive: bool, whether to only return spikes from the longest chunk
                                 (at least consecutive_n_seconds total).
                                 If False, all spikes belonging to good chunks are returned.
        - consecutive_n_seconds: float, minimum tolerated size (in seconds) of recording section
                                 with consecutive good chunks for its spikes to be returned.
                                 If less than consecutive_n_seconds good sections in total, does not pass.
        
    Returns:
        - train_filtered: array of spike times (in samples)
                          belonging to recording chunks where both FP and FN rates are low enough.
                          
    train_quality docstring:
    
    {train_quality.__doc__}
    """
    dp = Path(dp)
    t = trn(dp,unit)
    t_s=t/30000
    good_spikes_m, good_fp_start_end, good_fn_start_end = train_quality(dp, unit, period_m,
                    fp_chunk_span, fp_chunk_size, fn_chunk_span, fn_chunk_size, use_or_operator,
                    violations_ms, fp_threshold, fn_threshold, again, save, verbose, plot_debug)

    # use spike times themselves to define beginning and end of good Sections
    # as the FP and FN sections do not necessarily overlap
    edges = np.diff([0]+list(good_spikes_m.astype(int))+[0])
    good_left = np.nonzero(edges==1)[0]
    good_right = np.nonzero(edges==-1)[0]-1
    good_sections = [[t_s[l], t_s[r]] for l,r in zip(good_left, good_right)]
    
    if len(good_sections)>0:#
        total_good_sections = sum([s[1]-s[0] for s in good_sections])
        if use_consecutive:
            # get the longest consecutive section of time that passes
            # both our criteria
            maxrun = -1
            run_len = {}
            for x in good_sections:
                mrun = run_len[x] = run_len.get(x-1, 0) + 1
                if mrun > maxrun:
                    maxend, maxrun = x, mrun
            consecutive_good_chunk = list(range(maxend-maxrun+1, maxend+1))
            if len(consecutive_good_chunk) > consecutive_n_seconds:
                good_spikes_m = (t_s>consecutive_good_chunk[0])&(t_s<consecutive_good_chunk[-1]+1)
                return t[good_spikes_m], good_spikes_m
        else:
            if total_good_sections > consecutive_n_seconds:
                return t[good_spikes_m], good_spikes_m
        
    if verbose: print('No consecutive section passed the filters')
    return np.array([0]), (t*0).astype(bool)


def gaussian_cut(x, a, mu, sigma, x_cut):
    g = a * np.exp(-(x - mu) ** 2 / (2 * sigma ** 2))
    g[x < x_cut] = 0
    return g


def curve_fit_(x, num, p1):
    pop_t = curve_fit(gaussian_cut, x, num, p1, maxfev=10000)
    return pop_t

def ampli_fit_gaussian_cut(x, n_bins):
    # inputs: vector we want to estimate where the missing values start
    # inputs: number of bins
    # returns: ???

    # make inputs into numpy array
    a = np.asarray(x, dtype='float64')
    # get a histogram of the data, with the  number of entries in each bin
    # and the bin edges
    num, bins = np.histogram(a, bins=n_bins)
    # bin bottom bracket with the most entries
    # this can return more than one value for the mode
    mode_seed = bins[np.where(num == max(num))]
    #mode_seed = bins[np.argmax(num)]
    # find the bin width
    bin_steps = np.diff(bins[0:2])[0]
    #get the mean values of each bin
    x = bins[0:len(bins) - 1] + bin_steps / 2
    # get the value of the start of the first bin
    next_low_bin = x[0] - bin_steps
    #next_low = bins[0] - bin_steps/2

    # now we make more bins so they go all the way to 0
    add_points = np.arange(start =  0, stop=next_low_bin, step=bin_steps)
    #add_points = np.arange(start=next_low_bin, stop=0, step=-bin_steps)
    #add_points = np.flipud(add_points)
    # concatenate the new bin midpoints with the old ones
    x = np.concatenate([add_points, x])
    zeros = np.zeros((len(add_points), 1))
    zeros = zeros.reshape(len(zeros), )
    # concatenate the old number of bin elements with 0 for the new bins
    num = np.concatenate([zeros, num])

    # if there is  more than one mode of the  distribution, mean them
    if len(mode_seed) > 1:
        mode_seed = np.mean(mode_seed)

    # return: max, new mod, std for non nan values, and first percentile
    p0 = [np.max(num), mode_seed, np.nanstd(a), np.percentile(a, 1)]
    p0 = np.asarray(p0, dtype='float64')

    # Curve fit
    popt = curve_fit_(x, num, p0)
    p0 = popt[0]

    return x, p0


def gaussian_amp_est(x, n_bins):
#    breakpoint()
    try:
        x1, p0 = ampli_fit_gaussian_cut(x, n_bins)
        n_fit = gaussian_cut(x1, a=p0[0], mu=p0[1], sigma=p0[2], x_cut=p0[3])
        min_amp = p0[3]
        n_fit_no_cut = 0
        #n_fit_no_cut = gaussian_cut(x1, a=p0[0], mu=p0[1], sigma=p0[2], x_cut=0)
        percent_missing = round(100 * norm.cdf((min_amp - p0[1]) / p0[2]), 2)

    except RuntimeError:
        x1, p0, min_amp, n_fit, n_fit_no_cut, percent_missing = None, None, None, None, None, np.nan

    return x1, p0, min_amp, n_fit, n_fit_no_cut, percent_missing


def estimate_bins(x, rule):

    n = len(x)
    maxi = max(x)
    mini = min(x)

    # Freedman-Diaconis rule
    if rule == 'Fd':

        data = np.asarray(x, dtype=np.float_)
        iqr_ = iqr(data, scale=1, nan_policy="omit")
        n = data.size
        bw = (2 * iqr_) / np.power(n, 1 / 3)
        datmin= min(data)
        datmax = max(data)
        datrng = datmax - datmin
        bins = int(datrng/bw + 1)

        # q75, q25 = np.percentile(x, [75, 25])
        # iqr_ = q75 - q25
        # print('iqr', iqr_)
        # h = 2 * iqr_ * (n ** (-1/3))
        # print('h', h)
        # b = int(round((maxi-mini)/h, 0))

        return bins

    # Square-root choice
    elif rule == 'Sqrt':
        b = int(np.sqrt(n))
        return b



from  npyx.corr import acg
from npyx.merger import assert_multi, get_dataset_id, get_ds_table, get_source_dp_u
from npyx.plot import plot_fp_fn_rates



import traceback

import npyx
from npyx.inout import get_npix_sync
from npyx.gl import get_units, read_metadata
from npyx.spk_t import ids, trn, trn_filtered
from npyx.spk_wvf import wvf, wvf_dsmatch, get_peak_chan, templates
from npyx.corr import ccg
from npyx.plot import plot_ccg, plot_wvf, plot_raw

prefix = "\033[34;1m--- "
red_prefix = "\033[91;1m--- "
suffix = "\033[0m" 

def test_npyx(dp, raise_error=False):
    """
    Function for unit testing of npyx core functions.

    Parameters:
    - dp: path to Neuropixels data directory
          (must be compatible with phy, e.g. kilosort output)
    - raise_error: bool, whether to pause and raise error when test fails
                   (allows to enter pydebug interactive mode,
                   given that %pdb was set in notebook/ipython session)
    """
    print(f"{prefix}npyx version {npyx.__version__} unit testing initiated, on directory {dp}...{suffix}\n")

    test_function(read_metadata, raise_error, dp=dp)

    test_function(get_npix_sync, raise_error, dp=dp)

    units = test_function(get_units, raise_error, True, dp=dp)
    if units is None:
        raise ValueError("Something went really wrong, get_units did not run. Fix code and try again before testing next function.")
    u = units[-1]
    u1 = units[-3]

    test_function(ids, raise_error, dp=dp, unit=u, again=1)

    test_function(trn, raise_error, dp=dp, unit=u, again=1)

    test_function(trn_filtered, raise_error, dp=dp, unit=u, plot_debug=True, again=1)

    test_function(wvf, raise_error, dp=dp, u=u, again=1)

    test_function(wvf_dsmatch, raise_error, dp=dp, u=u, plot_debug=True, again=1, verbose=True)

    test_function(get_peak_chan, raise_error, dp=dp, unit=u, again=1)

    test_function(templates, raise_error, dp=dp, u=u)

    test_function(ccg, raise_error, dp=dp, U=[u,u1], bin_size=0.2, win_size=40, again=1)

    test_function(plot_wvf, raise_error, dp=dp, u=u, again=1, use_dsmatch=True)

    test_function(plot_ccg, raise_error, dp=dp, units=[u,u1], as_grid=True, again=1)

    test_function(plot_raw, raise_error, dp=dp, times=[0.1,0.15], channels=list(range(50)), again=1)


def test_function(fun, raise_error=False, ret=False, **kwargs):
    """
    Function to test a function with rich printed information.

    Parameters:
    - fun: function to test
    - ret: bool, whether to return output of fun
    - raise_error: bool, whether to pause and raise error when test fails
                   (allows to enter pydebug interactive mode,
                   given that %pdb was set in notebook/ipython session)
    - kwargs: parameters to fun
    """
    try:
        r = fun(**kwargs)
        print(f"{prefix}Successfully ran '{fun.__name__}' from {fun.__module__}.{suffix}")
        if ret: return r
    except Exception as err:
        print(f"{red_prefix}Failed to run '{fun.__name__}' from {fun.__module__} with the following error:{suffix}")
        e = traceback.format_exc()
        print(e)
        if raise_error:
            print(("\033[34;1mIf you wish to enter interactive debugging, "
            "make sure to have run the magic command %pdb in your notebook/ipython session.\033[0m"))
            raise FailedNpyxTest().with_traceback(err.__traceback__) from err

class FailedNpyxTest(Exception):
    pass